"""Dependency injection framework for Python."""

from that_depends import providers
from that_depends.container import BaseContainer
from that_depends.injection import Provide, inject
from that_depends.providers import container_context
from that_depends.providers.context_resources import (
    ContextScope,
    ContextScopes,
    fetch_context_item,
    fetch_context_item_by_type,
    get_current_scope,
)


__all__ = [
    "BaseContainer",
    "ContextScope",
    "ContextScopes",
    "Provide",
    "container_context",
    "fetch_context_item",
    "fetch_context_item_by_type",
    "get_current_scope",
    "inject",
    "providers",
]



import asyncio
import contextlib
import functools
import inspect
import re
import threading
import typing
import warnings
from contextlib import AsyncExitStack, ExitStack

from that_depends.container import BaseContainer
from that_depends.exceptions import TypeNotBoundError
from that_depends.meta import BaseContainerMeta
from that_depends.providers import AbstractProvider, ContextResource
from that_depends.providers.context_resources import ContextScope, ContextScopes, container_context
from that_depends.providers.mixin import ProviderWithArguments


class ContextProviderError(Exception):
    """Exception raised when a context provider is used where it is not expected."""


P = typing.ParamSpec("P")
T = typing.TypeVar("T")
_STRING_PROVIDER_PATTERN = re.compile(r"^([^.]+)\.([^.]+)(?:\.(.+))?$")
_INJECTION_WARNING_MESSAGE: typing.Final[str] = "Expected injection, but nothing found. Remove @inject decorator."
_PROVIDE_MESSAGE: typing.Final[str] = (
    "Use @Container.inject or @inject(container=Container) if you wish to use Provide()"
)


@typing.overload
def inject(func: typing.Callable[P, T]) -> typing.Callable[P, T]: ...


@typing.overload
def inject(
    *,
    scope: ContextScope | None = ContextScopes.INJECT,
    container: BaseContainerMeta | None = None,
    enter_scope: bool = False,
) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]: ...


def inject(  # noqa: C901
    func: typing.Callable[P, T] | None = None,
    scope: ContextScope | None = ContextScopes.INJECT,
    container: BaseContainerMeta | None = None,
    enter_scope: bool = False,
) -> typing.Callable[P, T] | typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
    """Mark a function for dependency injection.

    Args:
        func: function or generator function to be wrapped.
        scope: scope to initialize ContextResources for.
        container: container from which to resolve dependencies marked with `Provide()`.
        enter_scope: enter the provided scope.

    Returns:
        wrapped function.

    """
    if scope == ContextScopes.ANY:
        msg = f"{scope} is not allowed in inject decorator."
        raise ValueError(msg)
    if scope is None and enter_scope:
        msg = "enter_scope cannot be used with scope=None."
        raise ValueError(msg)

    def _inject(
        func: typing.Callable[P, T],
    ) -> typing.Callable[P, T]:
        if inspect.isasyncgenfunction(func):
            if enter_scope:
                msg = "enter_scope cannot be used with async generator functions."
                raise ValueError(msg)
            return typing.cast(typing.Callable[P, T], _inject_to_async_gen(func))
        if inspect.isgeneratorfunction(func):
            if enter_scope:
                msg = "enter_scope cannot be used with generator functions."
                raise ValueError(msg)
            return typing.cast(typing.Callable[P, T], _inject_to_sync_gen(func))
        if inspect.iscoroutinefunction(func):
            return typing.cast(typing.Callable[P, T], _inject_to_async(func))

        return _inject_to_sync(func)

    def _inject_to_sync_gen(
        gen: typing.Callable[P, typing.Generator[T, typing.Any, typing.Any]],
    ) -> typing.Callable[P, typing.Generator[T, typing.Any, typing.Any]]:
        @functools.wraps(gen)
        def inner(*args: P.args, **kwargs: P.kwargs) -> typing.Generator[T, typing.Any, typing.Any]:
            signature = inspect.signature(gen)
            injected, kwargs = _resolve_arguments_sync(signature, scope, container, None, *args, **kwargs)  # type: ignore[assignment]

            if not injected:
                warnings.warn(_INJECTION_WARNING_MESSAGE, RuntimeWarning, stacklevel=1)

            g = gen(*args, **kwargs)
            result = yield from g
            return result

        return inner

    def _inject_to_async_gen(
        gen: typing.Callable[P, typing.AsyncGenerator[T, typing.Any]],
    ) -> typing.Callable[P, typing.AsyncGenerator[T, typing.Any]]:
        @functools.wraps(gen)
        async def inner(*args: P.args, **kwargs: P.kwargs) -> typing.AsyncGenerator[T, typing.Any]:
            signature = inspect.signature(gen)

            injected, kwargs = await _resolve_arguments_async(signature, scope, container, None, *args, **kwargs)  # type: ignore[assignment]

            if not injected:
                warnings.warn(_INJECTION_WARNING_MESSAGE, RuntimeWarning, stacklevel=1)

            async for item in gen(*args, **kwargs):
                yield item

        return inner

    def _inject_to_async(
        func: typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]],
    ) -> typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]]:
        @functools.wraps(func)
        async def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            if enter_scope:
                async with container_context(scope=scope):
                    return await _resolve_async(func, None, container, *args, **kwargs)
            else:
                return await _resolve_async(func, scope, container, *args, **kwargs)

        return inner

    def _inject_to_sync(
        func: typing.Callable[P, T],
    ) -> typing.Callable[P, T]:
        @functools.wraps(func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> T:
            if enter_scope:
                with container_context(scope=scope):
                    return _resolve_sync(func, None, container, *args, **kwargs)
            else:
                return _resolve_sync(func, scope, container, *args, **kwargs)

        return inner

    if func is None:
        return _inject
    return _inject(func)


_SYNC_SIGNATURE_CACHE: dict[typing.Callable[..., typing.Any], inspect.Signature] = {}
_THREADING_LOCK = threading.Lock()


async def _resolve_arguments_async(
    signature: inspect.Signature,
    scope: ContextScope | None,
    container: BaseContainerMeta | None,
    stack: AsyncExitStack | None,
    *args: typing.Any,  # noqa: ANN401
    **kwargs: typing.Any,  # noqa: ANN401
) -> tuple[bool, dict[str, typing.Any]]:
    injected = False
    context_providers: set[AbstractProvider[typing.Any]] = set()
    params = list(signature.parameters.items())

    for i, (field_name, param) in enumerate(params):
        default = param.default

        if i < len(args) or field_name in kwargs:
            if isinstance(default, (AbstractProvider, StringProviderDefinition)):
                injected = True
            continue

        if isinstance(default, StringProviderDefinition):
            injected = True
            resolved_val = await _resolve_provider_with_scope_async(default.provider, scope, stack, context_providers)
            kwargs[field_name] = resolved_val
        elif isinstance(default, AbstractProvider):
            injected = True
            resolved_val = await _resolve_provider_with_scope_async(default, scope, stack, context_providers)
            kwargs[field_name] = resolved_val

        elif isinstance(default, _Provide):
            injected = True
            if container is None:
                raise RuntimeError(_PROVIDE_MESSAGE)
            try:
                provider = container.get_provider_for_type(signature.parameters[field_name].annotation)
            except TypeNotBoundError as e:
                msg = f"Type {signature.parameters[field_name].annotation} is not bound to a provider."
                raise RuntimeError(msg) from e
            kwargs[field_name] = await _resolve_provider_with_scope_async(provider, scope, stack, context_providers)
    return injected, kwargs


def _resolve_arguments_sync(
    signature: inspect.Signature,
    scope: ContextScope | None,
    container: BaseContainerMeta | None,
    stack: contextlib.ExitStack | None,
    *args: typing.Any,  # noqa: ANN401
    **kwargs: typing.Any,  # noqa: ANN401
) -> tuple[bool, dict[str, typing.Any]]:
    injected = False
    context_providers: set[AbstractProvider[typing.Any]] = set()
    for i, (field_name, param) in enumerate(signature.parameters.items()):
        default = param.default
        if i < len(args) or field_name in kwargs:
            if isinstance(default, (AbstractProvider, StringProviderDefinition, _Provide)):
                injected = True
            continue

        if isinstance(default, StringProviderDefinition):
            injected = True
            kwargs[field_name] = _resolve_provider_with_scope_sync(default.provider, scope, stack, context_providers)
        elif isinstance(default, AbstractProvider):
            injected = True
            kwargs[field_name] = _resolve_provider_with_scope_sync(default, scope, stack, context_providers)
        elif isinstance(default, _Provide):
            injected = True
            if container is None:
                raise RuntimeError(_PROVIDE_MESSAGE)
            try:
                provider = container.get_provider_for_type(signature.parameters[field_name].annotation)
            except TypeNotBoundError as e:
                msg = f"Type {signature.parameters[field_name].annotation} is not bound to a provider."
                raise RuntimeError(msg) from e
            kwargs[field_name] = _resolve_provider_with_scope_sync(provider, scope, stack, context_providers)

    return injected, kwargs


def _resolve_sync(
    func: typing.Callable[P, T],
    scope: ContextScope | None,
    container: BaseContainerMeta | None = None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    if func not in _SYNC_SIGNATURE_CACHE:
        with _THREADING_LOCK:
            _SYNC_SIGNATURE_CACHE[func] = inspect.signature(func)
    signature = _SYNC_SIGNATURE_CACHE[func]

    with ExitStack() as stack:
        injected, kwargs = _resolve_arguments_sync(signature, scope, container, stack, *args, **kwargs)  # type: ignore[assignment]

        if not injected:
            warnings.warn(_INJECTION_WARNING_MESSAGE, RuntimeWarning, stacklevel=1)

        return func(*args, **kwargs)


_SIGNATURE_CACHE: dict[
    typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, typing.Any]], inspect.Signature
] = {}

_ASYNCIO_LOCK = asyncio.Lock()


async def _resolve_async(
    func: typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]],
    scope: ContextScope | None,
    container: BaseContainerMeta | None = None,
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    if func not in _SIGNATURE_CACHE:
        async with _ASYNCIO_LOCK:
            _SIGNATURE_CACHE[func] = inspect.signature(func)
    signature = _SIGNATURE_CACHE[func]

    async with AsyncExitStack() as stack:
        injected, kwargs = await _resolve_arguments_async(signature, scope, container, stack, *args, **kwargs)  # type: ignore[assignment]
        if not injected:
            warnings.warn(_INJECTION_WARNING_MESSAGE, RuntimeWarning, stacklevel=1)

        return await func(*args, **kwargs)

    raise RuntimeError  # pragma: no cover # to prevent mypy issue


async def _resolve_provider_with_scope_async(
    provider: AbstractProvider[T],
    scope: ContextScope | None,
    stack: AsyncExitStack | None,
    providers: set[AbstractProvider[typing.Any]],
) -> T:
    """Resolve a provider with given scope and stack.

    Use `stack=None` to ensure ContextResource providers are not allowed.

    Args:
        provider: provider to resolve.
        scope: scope to resolve provider in.
        stack: stack to use for context resources.
        providers: providers traversed.

    Returns:
        resolved value for the provider.

    Raises:
        ContextProviderError: if the stack is None.

    """
    await _add_provider_to_stack_async(provider, stack, scope, providers)
    return await provider.resolve()


async def _add_provider_to_stack_async(
    provider: AbstractProvider[T],
    stack: AsyncExitStack | None,
    scope: ContextScope | None,
    providers: set[AbstractProvider[typing.Any]],
) -> None:
    if provider in providers:
        return
    providers.add(provider)

    if not scope:
        return
    if isinstance(provider, ProviderWithArguments):
        provider._register_arguments()  # noqa: SLF001

        parents = provider._parents  # noqa: SLF001
        for parent in parents:
            await _add_provider_to_stack_async(parent, stack, scope, providers)
    if isinstance(provider, ContextResource):
        provider_scope = provider.get_scope()
        if provider_scope in (ContextScopes.ANY, scope):
            if stack is None:
                msg = (
                    f"No stack exists, cannot initialize context for {provider} using scope {scope}.\n"
                    f"Note: @inject cannot initialize context for ContextResources when wrapping a generator."
                )
                raise ContextProviderError(msg)
            await stack.enter_async_context(provider.context_async(force=True))


def _resolve_provider_with_scope_sync(
    provider: AbstractProvider[T],
    scope: ContextScope | None,
    stack: ExitStack | None,
    providers: set[AbstractProvider[typing.Any]],
) -> T:
    _add_provider_to_stack_sync(provider, stack, scope, providers)
    return provider.resolve_sync()


def _add_provider_to_stack_sync(
    provider: AbstractProvider[T],
    stack: ExitStack | None,
    scope: ContextScope | None,
    providers: set[AbstractProvider[typing.Any]],
) -> None:
    if provider in providers:
        return
    providers.add(provider)

    if not scope:
        return
    if isinstance(provider, ProviderWithArguments):
        provider._register_arguments()  # noqa: SLF001

        parents = provider._parents  # noqa: SLF001
        for parent in parents:
            _add_provider_to_stack_sync(parent, stack, scope, providers)

    if isinstance(provider, ContextResource):
        provider_scope = provider.get_scope()
        if provider_scope in (ContextScopes.ANY, scope):
            if stack is None:
                msg = (
                    f"No stack exists, cannot initialize context for {provider} using scope {scope}.\n"
                    f"Note: @inject cannot initialize context for ContextResources when wrapping a generator."
                )
                raise ContextProviderError(msg)
            stack.enter_context(provider.context_sync(force=True))


class StringProviderDefinition:
    """Provider definition from a string."""

    def __init__(self, definition: str) -> None:
        """Initialize the provider definition.

        Args:
            definition: provider definition in the format container.provider[.attr1.attr2...].

        """
        self._definition = definition
        self._container_name, self._provider_name, self._attrs = self._validate_and_extract_provider_definition()

    def _validate_and_extract_provider_definition(self) -> tuple[str, str, list[str]]:
        match = re.match(_STRING_PROVIDER_PATTERN, self._definition)
        if match:
            container_name = match.group(1)
            provider_name = match.group(2)
            attrs = match.group(3).split(".") if match.group(3) else []
            return container_name, provider_name, attrs
        msg = f"Invalid provider definition: {self._definition}"
        raise ValueError(msg)

    def _get_container_by_name(self) -> type["BaseContainer"]:
        containers_in_scope = BaseContainerMeta.get_instances()
        try:
            return containers_in_scope[self._container_name]
        except KeyError as e:
            msg = f"Container {self._container_name} not found in scope!"
            raise ValueError(msg) from e

    def _get_provider_by_name(self) -> AbstractProvider[typing.Any]:
        container = self._get_container_by_name()
        try:
            provider = container.get_providers()[self._provider_name]
        except KeyError as e:
            msg = f"Provider {self._provider_name} not found in container {self._container_name}"
            raise ValueError(msg) from e
        for attr in self._attrs:
            provider = getattr(provider, attr)
        return provider

    @property
    def provider(self) -> AbstractProvider[typing.Any]:
        """Get the provider instance."""
        return self._get_provider_by_name()


class _Provide:
    def __getitem__(self, provider: AbstractProvider[T] | str) -> T | typing.Any:  # noqa: ANN401
        if isinstance(provider, str):
            return StringProviderDefinition(provider)  # will be resolved later
        return typing.cast(T, provider)

    def __call__(self) -> typing.Any:  # noqa: ANN401
        """Marker for automatic dependency injection."""
        return self


Provide: typing.Final[_Provide] = _Provide()



class TypeNotBoundError(Exception):
    """Exception raised when a type is not bound to a provider."""


class StateNotInitializedError(Exception):
    """Exception raised when the state is not initialized."""



import typing
from types import TracebackType
from typing import Any, Optional

from faststream import BaseMiddleware
from typing_extensions import override

from that_depends import container_context
from that_depends.providers.context_resources import ContextScope, SupportsContext
from that_depends.utils import UNSET, Unset, is_set


class DIContextMiddleware(BaseMiddleware):
    """Initializes the container context for faststream brokers."""

    def __init__(
        self,
        *context_items: SupportsContext[Any],
        global_context: dict[str, Any] | Unset = UNSET,
        scope: ContextScope | Unset = UNSET,
    ) -> None:
        """Initialize the container context middleware.

        Args:
            *context_items (SupportsContext[Any]): Context items to initialize.
            global_context (dict[str, Any] | Unset): Global context to initialize the container.
            scope (ContextScope | Unset): Context scope to initialize the container.

        """
        super().__init__()
        self._context: container_context | None = None
        self._context_items = set(context_items)
        self._global_context = global_context
        self._scope = scope

    @override
    async def on_receive(self) -> None:
        self._context = container_context(
            *self._context_items,
            scope=self._scope if is_set(self._scope) else None,
            global_context=self._global_context if is_set(self._global_context) else None,
        )
        await self._context.__aenter__()

    @override
    async def after_processed(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: Optional["TracebackType"] = None,
    ) -> bool | None:
        if self._context is not None:
            await self._context.__aexit__(exc_type, exc_val, exc_tb)
        return None

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> "DIContextMiddleware":  # noqa: ARG002, ANN401
        """Create an instance of DIContextMiddleware."""
        return DIContextMiddleware(*self._context_items, scope=self._scope, global_context=self._global_context)



import abc
import typing

from typing_extensions import override

from that_depends.providers.base import AbstractProvider
from that_depends.providers.mixin import ProviderWithArguments


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class AbstractFactory(ProviderWithArguments, AbstractProvider[T_co], abc.ABC):
    """Base class for all factories.

    This class defines the interface for factories that provide
    resources both synchronously and asynchronously.
    """

    @property
    def provider(self) -> typing.Callable[[], typing.Coroutine[typing.Any, typing.Any, T_co]]:
        """Returns an async provider function.

        The async provider function can be awaited to resolve the resource.

        Returns:
            Callable[[], Coroutine[Any, Any, T_co]]: The async provider function.

        Example:
            ```python
            provider = my_factory.provider
            resource = await provider()
            ```

        """
        return self.resolve

    @property
    def provider_sync(self) -> typing.Callable[[], T_co]:
        """Return a sync provider function.

        The sync provider function can be called to resolve the resource synchronously.

        Returns:
            Callable[[], T_co]: The sync provider function.

        Example:
            ```python
            provider = my_factory.provider_sync
            resource = provider()
            ```

        """
        return self.resolve_sync


class Factory(AbstractFactory[T_co]):
    """Provides an instance by calling a sync method.

    A typical usage scenario is to wrap a synchronous function
    that returns a resource. Each call to the provider or sync_provider
    produces a new instance of that resource.

    Example:
        ```python
        def build_resource(text: str, number: int):
            return f"{text}-{number}"

        factory = Factory(build_resource, "example", 42)
        resource = factory.provider_sync()  # "example-42"
        ```

    """

    def _register_arguments(self) -> None:
        self._register(self._args)
        self._register(self._kwargs.values())

    def _deregister_arguments(self) -> None:
        raise NotImplementedError

    __slots__ = "_args", "_factory", "_kwargs", "_override"

    def __init__(self, factory: typing.Callable[P, T_co], *args: P.args, **kwargs: P.kwargs) -> None:
        """Initialize a Factory instance.

        Args:
            factory (Callable[P, T_co]): Function that returns the resource.
            *args: Arguments to pass to the factory function.
            **kwargs: Keyword arguments to pass to the factory function.

        """
        super().__init__()
        self._factory: typing.Final = factory
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs
        self._register_arguments()

    @override
    async def resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        return self._factory(
            *[  # type: ignore[arg-type]
                await x.resolve() if isinstance(x, AbstractProvider) else x for x in self._args
            ],
            **{  # type: ignore[arg-type]
                k: await v.resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()
            },
        )

    @override
    def resolve_sync(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        return self._factory(
            *[  # type: ignore[arg-type]
                x.resolve_sync() if isinstance(x, AbstractProvider) else x for x in self._args
            ],
            **{  # type: ignore[arg-type]
                k: v.resolve_sync() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()
            },
        )


class AsyncFactory(AbstractFactory[T_co]):
    """Provides an instance by calling an async method.

    Similar to `Factory`, but requires an async function. Each call
    to the provider or `provider` property is awaited to produce a new instance.

    Example:
        ```python
        async def async_build_resource(text: str):
            await some_async_operation()
            return text.upper()

        async_factory = AsyncFactory(async_build_resource, "example")
        resource = await async_factory.provider()  # "EXAMPLE"
        ```

    """

    def _register_arguments(self) -> None:
        self._register(self._args)
        self._register(self._kwargs.values())

    def _deregister_arguments(self) -> None:
        raise NotImplementedError

    __slots__ = "_args", "_factory", "_kwargs", "_override"

    def __init__(self, factory: typing.Callable[P, typing.Awaitable[T_co]], *args: P.args, **kwargs: P.kwargs) -> None:
        """Initialize an AsyncFactory instance.

        Args:
            factory (Callable[P, Awaitable[T_co]]): Async function that returns the resource.
            *args: Arguments to pass to the factory function.
            **kwargs: Keyword arguments to pass to the factory


        """
        super().__init__()
        self._factory: typing.Final = factory
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs
        self._register_arguments()

    @override
    async def resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        return await self._factory(
            *[await x.resolve() if isinstance(x, AbstractProvider) else x for x in self._args],  # type: ignore[arg-type]
            **{k: await v.resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},  # type: ignore[arg-type]
        )

    @override
    def resolve_sync(self) -> typing.NoReturn:
        msg = "AsyncFactory cannot be resolved synchronously"
        raise RuntimeError(msg)



import typing

from typing_extensions import override

from that_depends.entities.resource_context import ResourceContext
from that_depends.providers.base import AbstractResource, ResourceCreatorType
from that_depends.providers.mixin import SupportsTeardown


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Resource(SupportsTeardown, AbstractResource[T_co]):
    """Provides a resource that is resolved once and cached for future usage.

    Unlike a singleton, this provider includes finalization logic and can be
    used with a generator or async generator to manage resource lifecycle.
    It also supports usage with `typing.ContextManager` or `typing.AsyncContextManager`.
    Threading and asyncio concurrency are supported, ensuring only one instance
    is created regardless of concurrent resolves.

    Example:
        ```python
        async def create_async_resource():
            try:
                yield "async resource"
            finally:
                # Finalize resource
                pass

        class MyContainer:
            async_resource = Resource(create_async_resource)

        async def main():
            async_resource_instance = await MyContainer.async_resource.async_resolve()
            await MyContainer.async_resource.tear_down()
        ```

    """

    __slots__ = (
        "_args",
        "_context",
        "_creator",
        "_creator",
        "_is_async",
        "_kwargs",
        "_override",
    )

    def __init__(
        self,
        creator: ResourceCreatorType[P, T_co],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Initialize the Resource provider with a callable for resource creation.

        The callable can be a generator or async generator that yields the resource
        (with optional teardown logic), or a context manager. Only one instance will be
        created and cached until explicitly torn down.

        Args:
            creator: The callable, generator, or context manager that creates the resource.
            *args: Positional arguments passed to the creator.
            **kwargs: Keyword arguments passed to the creator.

        Example:
            ```python
            def custom_creator(name: str):
                try:
                    yield f"Resource created for {name}"
                finally:
                    pass  # Teardown

            resource_provider = Resource(custom_creator, "example")
            instance = resource_provider.sync_resolve()
            resource_provider.tear_down()
            ```

        """
        super().__init__(creator, *args, **kwargs)
        self._context: typing.Final[ResourceContext[T_co]] = ResourceContext(is_async=self.is_async)

    def _fetch_context(self) -> ResourceContext[T_co]:
        return self._context

    @override
    async def tear_down(self, propagate: bool = True) -> None:
        """Tear down the resource if it has been created.

        If the resource was never resolved, or was already torn down,
        calling this method has no effect.

        Example:
            ```python
            # Assuming my_provider was previously resolved
            await my_provider.tear_down()
            ```

        """
        await self._fetch_context().tear_down()
        self._deregister_arguments()
        if propagate:
            await self._tear_down_children()

    @override
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        """Sync tear down the resource if it has been created.

        If the resource was never resolved, or was already torn down,
        calling this method has no effect.

        If you try to sync tear down an async resource, this will raise an exception.

        Example:
            ```python
            # Assuming my_provider was previously resolved
            my_provider.sync_tear_down()
            ```

        """
        self._fetch_context().tear_down_sync(propagate=propagate, raise_on_async=raise_on_async)
        self._deregister_arguments()
        if propagate:
            self._tear_down_children_sync(propagate=propagate, raise_on_async=raise_on_async)



import abc
import typing
import warnings
from collections.abc import MutableMapping
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from threading import Lock
from typing import TYPE_CHECKING

from typing_extensions import override

import that_depends
from that_depends.exceptions import TypeNotBoundError
from that_depends.providers import AbstractProvider, Resource
from that_depends.providers.context_resources import ContextResource, ContextScope, ContextScopes, SupportsContext
from that_depends.providers.mixin import SupportsTeardown


if TYPE_CHECKING:
    from that_depends import BaseContainer  # pragma: no cover


P = typing.ParamSpec("P")
T = typing.TypeVar("T")


class DefaultScopeNotDefinedError(Exception):
    """Exception raised when default_scope is not defined."""


class _ContainerMetaDict(dict[str, typing.Any]):
    """Implements custom logic for the container metaclass."""

    @override
    def __setitem__(self, key: str, value: typing.Any) -> None:
        if isinstance(value, ContextResource) and value.get_scope() == ContextScopes.ANY:
            try:
                default_scope = self.__getitem__("default_scope")
                super().__setitem__(key, value.with_config(default_scope))
            except KeyError as e:
                msg = "Explicitly define default_scope before defining ContextResource providers."
                raise DefaultScopeNotDefinedError(msg) from e
        else:
            super().__setitem__(key, value)


class BaseContainerMeta(SupportsContext[None], abc.ABCMeta):
    """Metaclass for BaseContainer."""

    @override
    def get_scope(cls) -> ContextScope | None:
        if scope := getattr(cls, "default_scope", None):
            return typing.cast(ContextScope | None, scope)
        return ContextScopes.ANY

    @asynccontextmanager
    @override
    async def context_async(cls, force: bool = False) -> typing.AsyncIterator[None]:
        async with AsyncExitStack() as stack:
            for container in cls.get_containers():
                await stack.enter_async_context(container.context_async(force=force))
            for provider in cls.get_providers().values():
                if isinstance(provider, ContextResource):
                    await stack.enter_async_context(provider.context_async(force=force))
            yield

    @contextmanager
    @override
    def context_sync(cls, force: bool = False) -> typing.Iterator[None]:
        with ExitStack() as stack:
            for container in cls.get_containers():
                stack.enter_context(container.context_sync(force=force))
            for provider in cls.get_providers().values():
                if isinstance(provider, ContextResource) and not provider._is_async:  # noqa: SLF001
                    stack.enter_context(provider.context_sync(force=force))
            yield

    @override
    def supports_context_sync(cls) -> bool:
        return True

    _instances: typing.ClassVar[dict[str, type["BaseContainer"]]] = {}

    _MUTABLE_ATTRS = (
        "__abstractmethods__",
        "__parameters__",
        "_abc_impl",
        "providers",
        "containers",
        "alias",
        "default_scope",
    )

    _lock: Lock = Lock()
    alias: str | None

    def __init__(cls, name: str, bases: tuple[type, ...], namespace: dict[str, typing.Any]) -> None:
        """Initialize the container class."""
        super().__init__(name, bases, namespace)
        cls_name = cls.name()
        with cls._lock:
            if name != "BaseContainer":
                if cls_name in cls._instances:
                    warnings.warn(f"Overwriting container '{cls_name}'", UserWarning, stacklevel=2)
                cls._instances[cls_name] = typing.cast(type["BaseContainer"], cls)

    def name(cls) -> str:
        """Get the name of the container class."""
        if alias := getattr(cls, "alias", None):
            return typing.cast(str, alias)
        return cls.__name__

    @classmethod
    @override
    def __prepare__(cls, name: str, bases: tuple[type, ...], /, **kwds: typing.Any) -> MutableMapping[str, object]:
        return _ContainerMetaDict()

    @override
    def __new__(cls, name: str, bases: tuple[type, ...], namespace: dict[str, typing.Any]) -> type:
        return super().__new__(cls, name, bases, namespace)

    @classmethod
    def get_instances(cls) -> dict[str, type["BaseContainer"]]:
        """Get all instances that inherit from BaseContainer."""
        return cls._instances

    @override
    def __setattr__(cls, key: str, value: typing.Any) -> None:
        if key in cls._MUTABLE_ATTRS:  # Allow modification of mutable attributes
            super().__setattr__(key, value)
        else:
            msg = f"Cannot add new attribute '{key}' to class '{cls.__name__}'"
            raise AttributeError(msg)

    def get_providers(cls) -> dict[str, "AbstractProvider[typing.Any]"]:
        """Get all connected providers."""
        if not hasattr(cls, "providers"):
            cls.providers = {k: v for k, v in cls.__dict__.items() if isinstance(v, AbstractProvider)}
        return cls.providers

    def get_containers(cls) -> list[type["BaseContainer"]]:
        """Get all connected containers."""
        if not hasattr(cls, "containers"):
            cls.containers: list[type[BaseContainer]] = []

        return cls.containers

    async def tear_down(cls) -> None:
        """Tear down all singleton and resource providers."""
        for provider in reversed(cls.get_providers().values()):
            if isinstance(provider, SupportsTeardown):
                await provider.tear_down()

        for container in cls.get_containers():
            await container.tear_down()

    def get_provider_for_type(cls, t: type[T]) -> AbstractProvider[T]:
        """Get a provider for a given type.

        Args:
            t: type to get the provider for.

        Returns:
            Provider for the given type.

        """
        for provider in cls.get_providers().values():
            if provider._has_contravariant_bindings:  # noqa: SLF001
                for bind in provider._bindings:  # noqa: SLF001
                    if issubclass(bind, t):
                        return provider
            elif t in provider._bindings:  # noqa: SLF001
                return provider
        msg = f"Type {t} is not bound to any provider in container {cls.name()}"
        raise TypeNotBoundError(msg)

    def tear_down_sync(cls) -> None:
        """Tear down all sync singleton and resource providers."""
        for provider in reversed(cls.get_providers().values()):
            if isinstance(provider, Resource):
                if not provider._is_async:  # noqa: SLF001
                    provider.tear_down_sync()
                continue

            if isinstance(provider, SupportsTeardown):
                provider.tear_down_sync()
                continue
        for container in cls.get_containers():
            container.tear_down_sync()

    @typing.overload
    def inject(cls, func: typing.Callable[P, T]) -> typing.Callable[P, T]: ...

    @typing.overload
    def inject(
        cls,
        *,
        scope: ContextScope | None = ContextScopes.INJECT,
        enter_scope: bool = False,
    ) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]: ...

    def inject(
        cls,
        func: typing.Callable[P, T] | None = None,
        scope: ContextScope | None = ContextScopes.INJECT,
        enter_scope: bool = False,
    ) -> typing.Callable[P, T] | typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
        """Inject dependencies from this container into a function.

        If your function does not use `Provide()` for dependency markers, consider using `@inject` instead.
        """
        return (
            that_depends.inject(scope=scope, container=cls, enter_scope=enter_scope)
            if func is None
            else that_depends.inject(scope=scope, container=cls, enter_scope=enter_scope)(func)
        )



from typing import Any, ClassVar, TypeGuard, TypeVar


T = TypeVar("T")


class _Singleton(type):
    _instances: ClassVar[dict[type, Any]] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Unset(metaclass=_Singleton):
    """Represents an unset value."""


UNSET = Unset()


def is_set(value: T | Unset) -> TypeGuard[T]:
    """Check if a value is set (not UNSET)."""
    return value is not UNSET



"""Selection based providers."""

import typing

from typing_extensions import override

from that_depends.providers.base import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


class Selector(AbstractProvider[T_co]):
    """Chooses a provider based on a key returned by a selector function.

    This class allows you to dynamically select and resolve one of several
    named providers at runtime. The provider key is determined by a
    user-supplied selector function.

    Examples:
        ```python
        def environment_selector():
            return "local"

        selector_instance = Selector(
            environment_selector,
            local=LocalStorageProvider(),
            remote=RemoteStorageProvider(),
        )

        # Synchronously resolve the selected provider
        service = selector_instance.sync_resolve()
        ```

    """

    __slots__ = "_override", "_providers", "_selector"

    def __init__(
        self, selector: typing.Callable[[], str] | AbstractProvider[str] | str, **providers: AbstractProvider[T_co]
    ) -> None:
        """Initialize a new Selector instance.

        Args:
            selector (Callable[[], str]): A function that returns the key
                of the provider to use.
            **providers (AbstractProvider[T_co]): The named providers from
                which one will be selected based on the `selector`.

        Examples:
            ```python
            def my_selector():
                return "remote"

            my_selector_instance = Selector(
                my_selector,
                local=LocalStorageProvider(),
                remote=RemoteStorageProvider(),
            )

            # The "remote" provider will be selected
            selected_service = my_selector_instance.sync_resolve()
            ```

        """
        super().__init__()
        self._selector: typing.Final[typing.Callable[[], str] | AbstractProvider[str] | str] = selector
        self._providers: typing.Final = providers

    @override
    async def resolve(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        if isinstance(self._selector, AbstractProvider):
            selected_key = await self._selector.resolve()
        else:
            selected_key = self._get_selected_key()

        self._validate_key(selected_key)

        return await self._providers[selected_key].resolve()

    @override
    def resolve_sync(self) -> T_co:
        if self._override:
            return typing.cast(T_co, self._override)

        if isinstance(self._selector, AbstractProvider):
            selected_key = self._selector.resolve_sync()
        else:
            selected_key = self._get_selected_key()

        self._validate_key(selected_key)

        return self._providers[selected_key].resolve_sync()

    def _get_selected_key(self) -> str:
        if callable(self._selector):
            selected_key = self._selector()
        elif isinstance(self._selector, str):
            selected_key = self._selector
        else:
            msg = f"Invalid selector type: {type(self._selector)}, expected str, or a provider/callable returning str"
            raise TypeError(msg)
        return typing.cast(str, selected_key)

    def _validate_key(self, key: str) -> None:
        if key not in self._providers:
            msg = f"Key '{key}' not found in providers"
            raise KeyError(msg)



import abc


class CannotTearDownSyncError(RuntimeError):
    """Raised when attempting to tear down an async resource in sync mode."""


class SupportsTeardown(abc.ABC):
    """Interface for objects that support teardown."""

    @abc.abstractmethod
    async def tear_down(self, propagate: bool = True) -> None:
        """Perform any necessary cleanup operations.

        This method is called when the object is no longer needed.
        """

    @abc.abstractmethod
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        """Perform any necessary cleanup operations.

        This method is called when the object is no longer needed.
        """


class ProviderWithArguments(abc.ABC):
    """Interface for providers that require arguments."""

    @abc.abstractmethod
    def _register_arguments(self) -> None:
        """Register arguments for the provider."""

    @abc.abstractmethod
    def _deregister_arguments(self) -> None:
        """Deregister arguments for the provider."""



import inspect
import typing
from contextlib import contextmanager
from typing import overload

from typing_extensions import override

from that_depends.meta import BaseContainerMeta
from that_depends.providers import AbstractProvider, AsyncSingleton, Resource, Singleton
from that_depends.providers.context_resources import ContextScope, ContextScopes


if typing.TYPE_CHECKING:
    import typing_extensions


T = typing.TypeVar("T")
P = typing.ParamSpec("P")


class BaseContainer(metaclass=BaseContainerMeta):
    """Base container class."""

    alias: str | None = None
    providers: dict[str, AbstractProvider[typing.Any]]
    containers: list[type["BaseContainer"]]
    default_scope: ContextScope | None = ContextScopes.ANY

    @classmethod
    @overload
    def context(cls, func: typing.Callable[P, T]) -> typing.Callable[P, T]: ...

    @classmethod
    @overload
    def context(cls, *, force: bool = False) -> typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]: ...

    @classmethod
    def context(
        cls, func: typing.Callable[P, T] | None = None, force: bool = False
    ) -> typing.Callable[P, T] | typing.Callable[[typing.Callable[P, T]], typing.Callable[P, T]]:
        """Wrap a function with this resources' context.

        Args:
            func: function to be wrapped.
            force: force context initialization, ignoring scope.

        Returns:
            wrapped function or wrapper if func is None.

        """

        def _wrapper(func: typing.Callable[P, T]) -> typing.Callable[P, T]:
            if inspect.iscoroutinefunction(func):

                async def _async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                    async with cls.context_async(force=force):
                        return await func(*args, **kwargs)  # type: ignore[no-any-return]

                return typing.cast(typing.Callable[P, T], _async_wrapper)

            def _sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                with cls.context_sync(force=force):
                    return func(*args, **kwargs)

            return _sync_wrapper

        if func:
            return _wrapper(func)
        return _wrapper

    @override
    def __new__(cls, *_: typing.Any, **__: typing.Any) -> "typing_extensions.Self":
        msg = f"{cls.__name__} should not be instantiated"
        raise RuntimeError(msg)

    @classmethod
    def connect_containers(cls, *containers: type["BaseContainer"]) -> None:
        """Connect containers.

        When `init_resources` and `tear_down` is called,
        same method of connected containers will also be called.
        """
        if not hasattr(cls, "containers"):
            cls.containers = []

        cls.containers.extend(containers)

    @classmethod
    async def init_resources(cls) -> None:
        """Initialize all resources."""
        for provider in cls.get_providers().values():
            if isinstance(provider, Resource | Singleton | AsyncSingleton):
                await provider.resolve()

        for container in cls.get_containers():
            await container.init_resources()

    @classmethod
    def reset_override_sync(cls) -> None:
        """Reset all provider overrides."""
        for v in cls.get_providers().values():
            v.reset_override_sync()

    @classmethod
    def resolver(cls, item: typing.Callable[P, T]) -> typing.Callable[[], typing.Awaitable[T]]:
        """Decorate a function to automatically resolve dependencies on call by name.

        Args:
            item: objects for which the dependencies should be resolved.

        Returns:
            Async wrapped callable with auto-injected dependencies.

        """

        async def _inner() -> T:
            return await cls.resolve(item)

        return _inner

    @classmethod
    async def resolve(cls, object_to_resolve: typing.Callable[..., T]) -> T:
        """Inject dependencies into an object automatically by name."""
        signature: typing.Final = inspect.signature(object_to_resolve)
        kwargs = {}
        providers: typing.Final = cls.get_providers()
        for field_name, field_value in signature.parameters.items():
            if field_value.default is not inspect.Parameter.empty or field_name in ("_", "__"):
                continue

            if field_name not in providers:
                msg = f"Provider is not found, {field_name=}"
                raise RuntimeError(msg)

            kwargs[field_name] = await providers[field_name].resolve()

        return object_to_resolve(**kwargs)

    @classmethod
    @contextmanager
    def override_providers_sync(cls, providers_for_overriding: dict[str, typing.Any]) -> typing.Iterator[None]:
        """Override several providers with mocks simultaneously.

        Args:
            providers_for_overriding: {provider_name: mock} dictionary.

        Returns:
            None

        """
        current_providers: typing.Final = cls.get_providers()
        current_provider_names: typing.Final = set(current_providers.keys())
        given_provider_names: typing.Final = set(providers_for_overriding.keys())

        for given_name in given_provider_names:
            if given_name not in current_provider_names:
                msg = f"Provider with name {given_name!r} not found"
                raise RuntimeError(msg)

        for provider_name, mock in providers_for_overriding.items():
            provider = current_providers[provider_name]
            provider.override_sync(mock)

        try:
            yield
        finally:
            for provider_name in providers_for_overriding:
                provider = current_providers[provider_name]
                provider.reset_override_sync()



import typing
import uuid
from contextlib import contextmanager
from contextvars import ContextVar
from typing import TypeVar

from typing_extensions import override

from that_depends.exceptions import StateNotInitializedError
from that_depends.providers import AbstractProvider
from that_depends.utils import UNSET, Unset, is_set


T = TypeVar("T", covariant=False)

_STATE_UNSET_ERROR_MESSAGE: typing.Final[str] = (
    "State has not been initialized.\n"
    "Please use `with provider.init(state):` to initialize the state before resolving it."
)


class State(AbstractProvider[T]):
    """Provides a value that can be passed into the provider at runtime."""

    def __init__(self) -> None:
        """Create a state provider."""
        super().__init__()

        self._state: ContextVar[T | Unset] = ContextVar(f"STATE_{uuid.uuid4()}", default=UNSET)

    @contextmanager
    def init(self, state: T) -> typing.Iterator[T]:
        """Set the state provider's value.

        Args:
            state: value to store.

        Returns:
            Iterator[T_co]: A context manager that yields the state value.

        """
        token = self._state.set(state)
        yield state
        self._state.reset(token)

    @override
    async def resolve(self) -> T:
        return self.resolve_sync()

    @override
    def resolve_sync(self) -> T:
        value = self._state.get()
        if is_set(value):
            return value
        raise StateNotInitializedError(_STATE_UNSET_ERROR_MESSAGE)



"""Resource providers."""

import asyncio
import threading
import typing

from typing_extensions import override

from that_depends.providers.base import AbstractProvider
from that_depends.providers.mixin import ProviderWithArguments, SupportsTeardown


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Singleton(ProviderWithArguments, SupportsTeardown, AbstractProvider[T_co]):
    """A provider that creates an instance once and caches it for subsequent injections.

    This provider is safe to use concurrently in both threading and asyncio contexts.
    On the first call to either ``sync_resolve()`` or ``async_resolve()``, the instance
    is created by calling the provided factory. All future calls return the cached instance.

    Example:
        ```python
        def my_factory() -> float:
            return 0.5

        singleton = Singleton(my_factory)
        value1 = singleton.sync_resolve()
        value2 = singleton.sync_resolve()
        assert value1 == value2
        ```

    """

    __slots__ = "_args", "_asyncio_lock", "_factory", "_instance", "_kwargs", "_override", "_threading_lock"

    def __init__(self, factory: typing.Callable[P, T_co], *args: P.args, **kwargs: P.kwargs) -> None:
        """Initialize the Singleton provider.

        Args:
            factory: A callable that produces the instance to be provided.
            *args: Positional arguments to pass to the factory.
            **kwargs: Keyword arguments to pass to the factory.

        """
        super().__init__()
        self._factory: typing.Final = factory
        self._instance: T_co | None = None
        self._asyncio_lock: typing.Final = asyncio.Lock()
        self._threading_lock: typing.Final = threading.Lock()
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs

    def _register_arguments(self) -> None:
        self._register(self._args)
        self._register(self._kwargs.values())

    def _deregister_arguments(self) -> None:
        self._deregister(self._args)
        self._deregister(self._kwargs.values())

    @override
    async def resolve(self) -> T_co:
        if self._override is not None:
            self._register_arguments()
            return typing.cast(T_co, self._override)

        # lock to prevent resolving several times
        async with self._asyncio_lock:
            if self._instance is not None:
                return self._instance
            self._register_arguments()
            self._instance = self._factory(
                *[await x.resolve() if isinstance(x, AbstractProvider) else x for x in self._args],  # type: ignore[arg-type]
                **{  # type: ignore[arg-type]
                    k: await v.resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()
                },
            )
            return self._instance

    @override
    def resolve_sync(self) -> T_co:
        if self._override is not None:
            self._register_arguments()
            return typing.cast(T_co, self._override)

        # lock to prevent resolving several times
        with self._threading_lock:
            if self._instance is not None:
                return self._instance
            self._register_arguments()
            self._instance = self._factory(
                *[x.resolve_sync() if isinstance(x, AbstractProvider) else x for x in self._args],  # type: ignore[arg-type]
                **{k: v.resolve_sync() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},  # type: ignore[arg-type]
            )
            return self._instance

    @override
    async def tear_down(self, propagate: bool = True) -> None:
        """Reset the cached instance.

        After calling this method, the next async_resolve call will recreate the instance.
        """
        if self._instance is not None:
            self._instance = None
        self._deregister_arguments()
        if propagate:
            await self._tear_down_children()

    @override
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        """Reset the cached instance.

        After calling this method, the next resolve call will recreate the instance.
        """
        if self._instance is not None:
            self._instance = None
        self._deregister_arguments()
        if propagate:
            self._tear_down_children_sync(propagate=propagate, raise_on_async=raise_on_async)


class AsyncSingleton(ProviderWithArguments, SupportsTeardown, AbstractProvider[T_co]):
    """A provider that creates an instance asynchronously and caches it for subsequent injections.

    This provider is safe to use concurrently in asyncio contexts. On the first call
    to ``async_resolve()``, the instance is created by awaiting the provided factory.
    All subsequent calls return the cached instance.

    Example:
        ```python
        async def my_async_factory() -> float:
            return 0.5

        async_singleton = AsyncSingleton(my_async_factory)
        value1 = await async_singleton.async_resolve()
        value2 = await async_singleton.async_resolve()
        assert value1 == value2
        ```

    """

    __slots__ = "_args", "_asyncio_lock", "_factory", "_instance", "_kwargs", "_override"

    def __init__(
        self,
        factory: typing.Callable[P, typing.Awaitable[T_co]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """Initialize the AsyncSingleton provider.

        Args:
            factory: The asynchronous callable used to create the instance.
            *args: Positional arguments to pass to the factory.
            **kwargs: Keyword arguments to pass to the factory.

        """
        super().__init__()
        self._factory: typing.Final[typing.Callable[P, typing.Awaitable[T_co]]] = factory
        self._instance: T_co | None = None
        self._asyncio_lock: typing.Final = asyncio.Lock()
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs

    def _register_arguments(self) -> None:
        self._register(self._args)
        self._register(self._kwargs.values())

    def _deregister_arguments(self) -> None:
        self._deregister(self._args)
        self._deregister(self._kwargs.values())

    @override
    async def resolve(self) -> T_co:
        if self._override is not None:
            return typing.cast(T_co, self._override)

        # lock to prevent resolving several times
        async with self._asyncio_lock:
            if self._instance is not None:
                return self._instance

            self._register_arguments()

            self._instance = await self._factory(
                *[await x.resolve() if isinstance(x, AbstractProvider) else x for x in self._args],  # type: ignore[arg-type]
                **{  # type: ignore[arg-type]
                    k: await v.resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()
                },
            )
            return self._instance

    @override
    def resolve_sync(self) -> typing.NoReturn:
        msg = "AsyncSingleton cannot be resolved in an sync context."
        raise RuntimeError(msg)

    @override
    async def tear_down(self, propagate: bool = True) -> None:
        """Reset the cached instance.

        After calling this method, the next call to ``async_resolve()`` will recreate the instance.
        """
        if self._instance is not None:
            self._instance = None
        self._deregister_arguments()
        if propagate:
            await self._tear_down_children()

    @override
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        """Reset the cached instance.

        After calling this method, the next call to ``sync_resolve()`` will recreate the instance.
        """
        if self._instance is not None:
            self._instance = None
        self._deregister_arguments()
        if propagate:
            self._tear_down_children_sync(propagate=propagate, raise_on_async=raise_on_async)



import typing

from typing_extensions import override

from that_depends.providers.base import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


class List(AbstractProvider[list[T_co]]):
    """Provides multiple resources as a list.

    The `List` provider resolves multiple dependencies into a list.

    Example:
        ```python
        from that_depends import providers

        provider1 = providers.Factory(lambda: 1)
        provider2 = providers.Factory(lambda: 2)

        list_provider = List(provider1, provider2)

        # Synchronous resolution
        resolved_list = list_provider.resolve_sync()
        print(resolved_list)  # Output: [1, 2]

        # Asynchronous resolution
        import asyncio
        resolved_list_async = asyncio.run(list_provider.resolve())
        print(resolved_list_async)  # Output: [1, 2]
        ```

    """

    __slots__ = ("_providers",)

    def __init__(self, *providers: AbstractProvider[T_co]) -> None:
        """Create a new List provider instance.

        Args:
            *providers: List of providers to resolve.

        """
        super().__init__()
        self._providers: typing.Final = providers
        self._register(self._providers)

    @override
    def __getattr__(self, attr_name: str) -> typing.Any:
        msg = f"'{type(self)}' object has no attribute '{attr_name}'"
        raise AttributeError(msg)

    @override
    async def resolve(self) -> list[T_co]:
        return [await x.resolve() for x in self._providers]

    @override
    def resolve_sync(self) -> list[T_co]:
        return [x.resolve_sync() for x in self._providers]

    @override
    async def __call__(self) -> list[T_co]:
        return await self.resolve()


class Dict(AbstractProvider[dict[str, T_co]]):
    """Provides multiple resources as a dictionary.

    The `Dict` provider resolves multiple named dependencies into a dictionary.

    Example:
        ```python
        from that_depends import providers

        provider1 = providers.Factory(lambda: 1)
        provider2 = providers.Factory(lambda: 2)

        dict_provider = Dict(key1=provider1, key2=provider2)

        # Synchronous resolution
        resolved_dict = dict_provider.sync_resolve()
        print(resolved_dict)  # Output: {"key1": 1, "key2": 2}

        # Asynchronous resolution
        import asyncio
        resolved_dict_async = asyncio.run(dict_provider.async_resolve())
        print(resolved_dict_async)  # Output: {"key1": 1, "key2": 2}
        ```

    """

    __slots__ = ("_providers",)

    def __init__(self, **providers: AbstractProvider[T_co]) -> None:
        """Create a new Dict provider instance.

        Args:
            **providers: Dictionary of providers to resolve.

        """
        super().__init__()
        self._providers: typing.Final = providers
        self._register(self._providers.values())

    @override
    def __getattr__(self, attr_name: str) -> typing.Any:
        msg = f"'{type(self)}' object has no attribute '{attr_name}'"
        raise AttributeError(msg)

    @override
    async def resolve(self) -> dict[str, T_co]:
        return {key: await provider.resolve() for key, provider in self._providers.items()}

    @override
    def resolve_sync(self) -> dict[str, T_co]:
        return {key: provider.resolve_sync() for key, provider in self._providers.items()}



import typing
from collections.abc import Callable
from inspect import signature

from fastapi import Depends, Request, Response
from fastapi.routing import APIRoute

from that_depends.injection import StringProviderDefinition
from that_depends.providers import AbstractProvider
from that_depends.providers.context_resources import ContextScope, SupportsContext, container_context


P = typing.ParamSpec("P")
T = typing.TypeVar("T")


def _adjust_fastapi_endpoint(endpoint: Callable[P, T]) -> Callable[P, T]:
    hints = typing.get_type_hints(endpoint, include_extras=True)
    sig = signature(endpoint)
    new_params = []
    for name, param in sig.parameters.items():
        if isinstance(param.default, AbstractProvider):
            hints[name] = Depends(param.default)
            new_params.append(param.replace(default=Depends(param.default)))
        elif isinstance(param.default, StringProviderDefinition):
            provider = param.default.provider
            hints[name] = Depends(provider)
            new_params.append(param.replace(default=Depends(provider)))
        else:
            new_params.append(param)
    endpoint.__annotations__ = hints
    endpoint.__signature__ = sig.replace(parameters=new_params)  # type: ignore[attr-defined]
    return endpoint


def create_fastapi_route_class(
    *context_items: SupportsContext[typing.Any],
    global_context: dict[str, typing.Any] | None = None,
    scope: ContextScope | None = None,
) -> type[APIRoute]:
    """Create a `that-depends` fastapi route class.

    Args:
        *context_items: Items to initialize context for.
        global_context: Global context to use.
        scope: scope to enter before on request.

    Returns:
        type[APIRoute]: A custom fastapi route class.

    """

    class _Route(APIRoute):
        """Custom that-depends router for FastAPI."""

        def __init__(self, path: str, endpoint: Callable[..., typing.Any], **kwargs: typing.Any) -> None:  # noqa: ANN401
            endpoint = _adjust_fastapi_endpoint(endpoint)
            super().__init__(path=path, endpoint=endpoint, **kwargs)

        def get_route_handler(self) -> Callable[[Request], typing.Coroutine[typing.Any, typing.Any, Response]]:
            original_route_handler = super().get_route_handler()

            async def _custom_route_handler(request: Request) -> Response:
                async with container_context(*context_items, global_context=global_context, scope=scope):
                    response: Response = await original_route_handler(request)
                    return response

            return _custom_route_handler if context_items or global_context or scope else original_route_handler

    return _Route
